from pathlib import Path
import tempfile

import invoke
import pytest

from mloq.api import setup_project, setup_requirements
from mloq.files import test_req


@pytest.fixture()
def repo_path():
    return Path(__file__).parent / "test_project"


def test_init_repository(project_config, template):
    with tempfile.TemporaryDirectory() as tmp:
        setup_project(
            path=Path(tmp),
            project_config=project_config,
            template=template,
            override=False,
        )


def test_setup_requirements():
    try:
        setup_requirements(test_req.src.parent, project_config={}, install=("all",))
    except invoke.exceptions.UnexpectedExit:
        pass
    setup_requirements(test_req.src.parent, project_config={}, install=("lint",))



"""Setup Git repository for the project."""


from pathlib import Path
import subprocess
from typing import Union

from mloq.config import Config
from mloq.failure import Failure


def setup_git(
    path: Union[Path, str],
    project_config: Config,
    template: Config,
) -> None:
    """Initialize a Git repository over the generated files."""
    git_init = project_config["git_init"]
    if not git_init:
        return
    message = template["git_message"]
    push = project_config["git_push"]
    branch = template["default_branch"]
    project_name = template["project_name"]
    owner = template["owner"]
    sign_off = project_config["open_source"]
    path = str(path)
    try:
        _git_cmd(path, "init")
        _git_cmd(path, *f"remote add origin ssh://git@github.com/{owner}/{project_name}".split())
        _git_cmd(path, *"add .".split())
        _git_cmd(path, *f"commit {'--signoff' if sign_off else ''} -m".split(), message)
        if push:
            _git_cmd(path, *f"push origin {branch}".split())
    except subprocess.CalledProcessError as e:
        raise Failure() from e


def _git_cmd(git_dir: str, *parts: str) -> None:
    subprocess.run(("git",) + parts, check=True, cwd=git_dir)



"""This module defines common functionality for dealing with project requirements."""
import os
from pathlib import Path
from typing import Iterable, Optional, Union

from invoke import run

from mloq.config.params import Config, is_empty
from mloq.directories import copy_file
from mloq.files import (
    data_science_req,
    data_viz_req,
    dogfood_req,
    File,
    lint_req,
    pytorch_req,
    tensorflow_req,
    test_req,
)


REQUIREMENTS_ALIASES = {
    data_science_req: ["data-science", "datascience", "ds"],
    pytorch_req: ["pytorch", "torch"],
    tensorflow_req: ["tensorflow", "tf"],
    data_viz_req: ["data-visualization", "data-viz", "data-vis", "dataviz", "datavis"],
}


def require_cuda(project_config: Optional[Config] = None) -> bool:
    """Return True if any of the project dependencies require CUDA."""
    project_config = project_config or {}
    if "requirements" not in project_config:
        return False
    options = project_config["requirements"]
    if is_empty(options):
        return False
    elif isinstance(options, str):
        options = [options]
    tf_alias = REQUIREMENTS_ALIASES[tensorflow_req]
    torch_alias = REQUIREMENTS_ALIASES[pytorch_req]
    for option in options:
        if option in tf_alias or option in torch_alias:
            return True
    return False


def get_aliased_requirements_file(option: str) -> File:
    """Get requirement file from aliased name."""
    for file, valid_alias in REQUIREMENTS_ALIASES.items():
        if option in valid_alias:
            return file
    if option == "dogfood":
        return dogfood_req
    raise KeyError(f"{option} is not a valid name. Valid aliases are {REQUIREMENTS_ALIASES}")


def read_requirements_file(option: str) -> str:
    """Return the content of the target requirements file form an aliased name."""
    req_file = get_aliased_requirements_file(option)
    with open(req_file.src, "r") as f:
        return f.read()


def compose_requirements(options: Iterable[str]) -> str:
    """
    Return the content requirements.txt file with pinned dependencies.

    The returned string contains the combined dependencies\
     for the different options sorted alphabetically.

    Args:
        options: Iterable containing the aliased names of the target dependencies for the project.

    Returns:
        str containing the pinned versions of all the selected requirements.
    """
    requirements_text = ""
    for i, opt in enumerate(options):
        pref = "\n" if i > 0 else ""  # Ensure one requirement per line
        requirements_text += pref + read_requirements_file(opt)
    # Sort requirements alphabetically
    requirements_text = "\n".join(sorted(requirements_text.split("\n"))).lstrip("\n")
    return requirements_text


def write_project_requirements(
    options: Iterable[str],
    out_path: Optional[Union[Path, str]] = None,
    out_name: str = "requirements.txt",
    override: bool = False,
):
    """
    Write the composed requirements.txt file.

    The writen file contains pinned dependencies for the libraries \
    required by the provided options.
    """
    if is_empty(set(options)):
        return
    out_path = os.getcwd() if out_path is None else out_path
    file_path = out_path / out_name
    if not override and file_path.exists():
        return
    requirements = compose_requirements(options) if options is not None else ""
    with open(file_path, "w") as f:
        f.write(requirements)


def write_dev_requirements(
    out_path: Optional[Union[Path, str]] = None,
    override: bool = False,
    test: bool = True,
    lint: bool = True,
):
    """Write requirements-lint.txt and requirements-test.txt in the target directory."""
    out_path = Path(os.getcwd() if out_path is None else out_path)
    if lint:
        copy_file(lint_req, out_path, override)
    if test:
        copy_file(test_req, out_path, override)


def write_requirements(
    out_path: Union[str, Path],
    options: Optional[Iterable[str]] = None,
    out_name="requirements.txt",
    override: bool = False,
    lint: bool = True,
    test: bool = True,
):
    """Write the different requirements.txt files for the project."""
    out_path = Path(out_path)
    if options is not None and not is_empty(options):
        write_project_requirements(
            options=options,
            out_path=out_path,
            out_name=out_name,
            override=override,
        )
    write_dev_requirements(out_path, override=override, test=test, lint=lint)


def install_requirement_file(path: Union[Path, str], py3: bool = True):
    """Install the dependencies listed in the target requirements file."""
    python = "python3" if py3 else "python"
    return run(f"{python} -m pip install -r {str(path)}")


def install_requirements(
    path: Union[Path, str],
    requirements: Optional[Union[str, bool, Path]] = None,
    test: Optional[Union[str, bool, Path]] = None,
    lint: Optional[Union[str, bool, Path]] = None,
    py3: bool = True,
):
    """Install the dependencies listed in the target requirements files."""
    path = Path(path)
    if requirements is not None and requirements:
        requirements = "requirements.txt" if isinstance(requirements, bool) else requirements
        requirements = requirements if isinstance(requirements, Path) else path / requirements
        install_requirement_file(requirements, py3=py3)
    if test is not None and test:
        test = "requirements-test.txt" if isinstance(test, bool) else test
        test = test if isinstance(test, Path) else path / test
        install_requirement_file(test, py3=py3)
    if lint is not None and lint:
        lint = "requirements-lint.txt" if isinstance(lint, bool) else lint
        lint = lint if isinstance(lint, Path) else path / lint
        install_requirement_file(lint, py3=py3)



"""The api module exposes the mloq functionality to the CLI."""
import copy
from pathlib import Path
from typing import List, Tuple, Union

from omegaconf import OmegaConf

from mloq.config import Config
from mloq.directories import create_project_directories
from mloq.files import mlproject, OPEN_SOURCE_FILES, ROOT_PATH_FILES, SCRIPTS, setup_py
from mloq.git import setup_git
from mloq.requirements import install_requirements, write_requirements
from mloq.templating import write_template
from mloq.workflows import setup_push_workflow


def setup_requirements(
    path: Union[Path, str],
    project_config: Config,
    lint: bool = True,
    test: bool = True,
    install: Union[str, Tuple[str], List[str], None] = None,
    override: bool = False,
) -> None:
    """Write requirements files and install them if requested."""
    if isinstance(install, (tuple, list)) and "all" in install or install == "all":
        install_reqs = install_lint = install_test = True
    elif isinstance(install, (tuple, list)):
        install_reqs = "reqs" in install or "requirements" in install
        install_lint = "lint" in install
        install_test = "test" in install
    else:
        install_reqs = install_lint = install_test = False
    options = project_config.get("requirements", ["empty"])
    write_requirements(
        out_path=path,
        options=options,
        test=test,
        lint=lint,
        override=override,
    )
    install_requirements(
        path=path,
        requirements=install_reqs,
        test=install_test,
        lint=install_lint,
        py3=True,
    )


def setup_project_files(
    path: Union[Path, str],
    template: Config,
    project_config: Config,
    override: bool = False,
) -> None:
    """Write the template for common repository config files."""
    path = Path(path)
    _template = copy.deepcopy(template)
    _template["project_name"] = _template["project_name"].replace("-", "_")
    project_name = _template["project_name"]
    create_project_directories(project_name=project_name, root_path=path, override=override)
    setup_root_files(
        template=_template,
        project_config=project_config,
        path=path,
        override=override,
    )


def setup_scripts(path: Union[str, Path], template: Config, override: bool = False) -> None:
    """Initialize CI scripts folder files."""
    path = Path(path)
    path = path if path.name == "scripts" else path / "scripts"
    for file in SCRIPTS:
        write_template(file, template=template, path=path, override=override)


def setup_root_files(
    path: Union[str, Path],
    template: Config,
    project_config: Config,
    override: bool = False,
) -> None:
    """Initialize root folder files."""
    for file in ROOT_PATH_FILES:
        write_template(file, template=template, path=path, override=override)
    if project_config.get("mlflow"):
        write_template(mlproject, template=template, path=path, override=override)
    if project_config.get("open_source"):
        for file in OPEN_SOURCE_FILES:
            write_template(file, template=template, path=path, override=override)
    propietary_classif = "License :: Other/Proprietary License"
    license_classifiers = {"mit": "License :: OSI Approved :: MIT License"}
    license_classif = license_classifiers.get(template["license"], propietary_classif)
    setup_template = {k: copy.deepcopy(v) for k, v in template.items()}
    setup_template.update({"license_classifier": license_classif})
    setup_template = OmegaConf.create(setup_template)
    write_template(setup_py, template=setup_template, path=path, override=override)


def setup_project(
    path: Union[str, Path],
    template: Config,
    project_config: Config,
    override: bool = False,
) -> None:
    """Initialize the project folder structure and all the filled in boilerplate files."""
    path = Path(path)
    setup_project_files(
        path=path,
        template=template,
        project_config=project_config,
        override=override,
    )
    setup_push_workflow(
        path=path,
        project_config=project_config,
        template=template,
        override=override,
    )
    setup_requirements(
        path=path,
        project_config=project_config,
        test=True,
        lint=True,
        override=override,
    )
    setup_scripts(
        path=path,
        template=template,
        override=override,
    )
    setup_git(
        path=path,
        project_config=project_config,
        template=template,
    )



"""config contains all the logic needed to define the configuration of a new mloq project."""
from typing import List, Set, Tuple, Union

from omegaconf import DictConfig


Config = DictConfig
Choices = Union[List[str], Tuple[str], Set[str]]



from pathlib import Path
import tempfile

from click.testing import CliRunner

from mloq.cli.main import cli
from mloq.tests.cli.cli_inputs import cli_input


def test_setup_interactive(cli_input):
    with tempfile.TemporaryDirectory() as tmp:
        runner = CliRunner(echo_stdin=False)
        result = runner.invoke(cli, ["setup", "-i", str(tmp)], input=cli_input)
    assert result.exit_code == 0, result.stdout


def test_setup_non_interactive():
    mloq_file = Path(__file__).parent.parent / "mloq.yml"
    with tempfile.TemporaryDirectory() as tmp:
        runner = CliRunner(echo_stdin=False)
        result = runner.invoke(cli, ["setup", "-f", str(mloq_file), str(tmp)])
        _ = runner.invoke(cli, ["setup", "-f", str(mloq_file), str(tmp)])
    assert result.exit_code == 0, result.stdout


def test_setup_error():
    mloq_file = Path(__file__).parent.parent / "mloq.yml"
    runner = CliRunner(echo_stdin=False)
    result = runner.invoke(cli, ["setup", "-f", str(mloq_file), "/"])
    msg = "Failed to setup the project:"
    fails = msg in result.output or result.exit_code == 1
    fails = fails or result.stderr_bytes is None  # Pass when running inside docker
    assert fails, result.output



"""This module defines common functionality for rendering and writing File templates."""
from datetime import datetime
from pathlib import Path
from typing import Union

from jinja2 import Environment, FileSystemLoader, select_autoescape

from mloq import _logger
from mloq.config import Config
from mloq.directories import read_file
from mloq.files import File, TEMPLATES_PATH, WORKFLOWS_PATH


jinja_env = Environment(
    loader=FileSystemLoader([str(TEMPLATES_PATH), str(WORKFLOWS_PATH)]),
    autoescape=select_autoescape(["html", "xml"]),
)


def render_template(file: File, template: Config) -> str:
    """
    Render a jinja template with the provided parameter dict.

    Args:
        file: File object representing the jinja template that will be rendered.
        template: Dictionary containing the parameters key and corresponding values \
                that will be used to render the template.

    Returns:
        String containing the rendered template.
    """
    if file.is_static:
        return read_file(file)
    jinja_template = jinja_env.get_template(str(file.name))
    jinja_template.globals["now"] = datetime.now
    return jinja_template.render(**template)


def write_template(file: File, template: Config, path: Union[Path, str], override: bool = False):
    """
    Create new file containing the rendered template found in source_path.

    Args:
        file: File object representing the jinja template that will be rendered.
        template: Dictionary containing the parameters key and corresponding values \
                that will be used to render the templates.
        path: Absolute path to the folder containing the target templates.
        override: If False, copy the file if it does not already exists in the \
                  target path. If True, overwrite the target file if it is already present.

    Returns:
        None.
    """
    path = Path(path)
    if not override and (path / file.dst).exists():
        _logger.debug(f"file {file.name} already exists. Skipping")
        return
    rendered = render_template(file, template)
    with open(path / file.dst, "w") as f:
        f.write(rendered)



import tempfile

from mloq.workflows import setup_push_workflow, setup_workflow_template


def test_setup_workflow_template():
    # test empty workflow
    with tempfile.TemporaryDirectory() as tmp:
        setup_workflow_template(root_path=tmp, project_config=dict(), template=dict())
    # Test invalid name workflow
    setup_workflow_template(root_path=tmp, project_config=dict(ci="miau_db"), template=dict())


def test_setup_push_workflow():
    assert setup_push_workflow(None, project_config={}, template={}, override=False) is None



"""Define Failure - the exception to raise when we break."""


class Failure(Exception):
    """Raised when a project setup critical error happens."""

    def __str__(self) -> str:
        """Delegate __str__ to the underlying cause if it exists."""
        if self.args or not self.__cause__:
            return super().__str__()
        return f"{type(self.__cause__).__name__}: {self.__cause__}"



"""This module defines common functionality for dealing with directories and files."""
import os
from pathlib import Path
from shutil import copyfile
from typing import Union

from mloq import _logger
from mloq.config import Config
from mloq.failure import Failure
from mloq.files import File, init, main, version


def read_file(file: File) -> str:
    """Return and string with the content of the provided file."""
    with open(file.src, "r") as f:
        return f.read()


def copy_file(file: File, path: Union[Path, str], override: bool = False) -> None:
    """
    Copy the file from src into dst.

    Args:
        file: File object representing the file that will be copied.
        path: Path to the destination of the copied file.
        override: If False, copy the file if it does not already exists in the \
                  target path. If True, overwrite the target file if it is already present.
    Returns:
        None.
    """
    target = path / file.dst
    if not os.path.isfile(str(target)) or override:
        copyfile(file.src, target)
    else:
        _logger.debug(f"file {file.name} already exists in {target}")


def create_project_directories(
    project_name: str,
    root_path: Path,
    project_config: Config = None,
    override: bool = False,
) -> None:
    """
    Initialize the folder structure of a new Python project.

    It creates a directory for the project containing an empty __init__.py file, and a \
    tests folder inside it with its corresponding __init__.py file. This function will only \
    create new files and folders in case they don't already exist, but it won't overwrite any \
    existing file.

    Args:
        project_name: Name of the project. This is the name of the directory that will be created.
        root_path: Absolute path where the new project folder will be created.
        project_config: Contains all the parameters that define how the project will be set up.
        override: If False, copy the file if it does not already exists in the \
                  target path. If True, overwrite the target file if it is already present.

    Returns:
        None.
    """
    project_config = project_config or {}
    # Project dir
    try:
        project_path = root_path / project_name
        os.makedirs(project_path, exist_ok=True)
        copy_file(init, project_path, override)
        copy_file(version, project_path, override)
        copy_file(main, project_path, override)
        # Test dir inside project
        test_path = project_path / "tests"
        os.makedirs(test_path, exist_ok=True)
        copy_file(init, test_path, override)
        # Scripts dir
        scripts_path = root_path / "scripts"
        os.makedirs(scripts_path, exist_ok=True)
    except (PermissionError, FileNotFoundError) as e:
        raise Failure() from e


def create_github_actions_directories(root_path: Union[str, Path]) -> None:
    """
    Initialize the folder structure for using GitHub actions workflows.

    It creates a .github directory on the target path, with an empty workflows \
    directory inside it, so the .github/workflows directory is ready to add GitHub \
    actions workflows. This function will only create new folders in case they don't already exist.

    Args:
        root_path: Absolute path where the .github folder will be crated.

    Returns:
        None.
    """
    gha_path = Path(root_path) / ".github"
    os.makedirs(gha_path, exist_ok=True)
    workflows_path = gha_path / "workflows"
    os.makedirs(workflows_path, exist_ok=True)



"""This module defines all the different assets accessible from mloq."""
from collections import namedtuple
from pathlib import Path
from typing import Optional, Union


File = namedtuple("File", "name src dst is_static")


def file(
    name: str,
    path: Union[Path, str],
    dst: Optional[Union[Path, str]] = None,
    is_static: bool = False,
):
    """Define a new asset as a File namedtuple."""
    dst = dst if dst is not None else name
    return File(name=name, src=Path(path) / name, dst=dst, is_static=is_static)


# Assets paths
ASSETS_PATH = Path(__file__).parent / "assets"
TEMPLATES_PATH = ASSETS_PATH / "templates"
STATIC_FILES_PATH = ASSETS_PATH / "static"
REQUIREMENTS_PATH = ASSETS_PATH / "requirements"
WORKFLOWS_PATH = ASSETS_PATH / "workflows"

# Common files
mloq_yml = file("mloq.yml", STATIC_FILES_PATH, is_static=True)
gitignore = file(".gitignore", STATIC_FILES_PATH, is_static=True)
dco = file("DCO.md", STATIC_FILES_PATH, is_static=True)
init = file("init.txt", STATIC_FILES_PATH, "__init__.py", is_static=True)
main = file("main.txt", STATIC_FILES_PATH, "__main__.py", is_static=True)
version = file("version.txt", STATIC_FILES_PATH, "version.py", is_static=True)
dockerfile_aarch64 = file("Dockerfile_aarch64", STATIC_FILES_PATH, is_static=True)
build_manylinux_sh = file("build-manylinux-wheels.sh", STATIC_FILES_PATH, is_static=True)
code_of_conduct = file("CODE_OF_CONDUCT.md", TEMPLATES_PATH)

# Requirements files
data_science_req = file("data-science.txt", REQUIREMENTS_PATH)
data_viz_req = file("data-visualization.txt", REQUIREMENTS_PATH, is_static=True)
pytorch_req = file("pytorch.txt", REQUIREMENTS_PATH, is_static=True)
tensorflow_req = file("tensorflow.txt", REQUIREMENTS_PATH, is_static=True)
lint_req = file("requirements-lint.txt", REQUIREMENTS_PATH, is_static=True)
test_req = file("requirements-test.txt", REQUIREMENTS_PATH, is_static=True)
dogfood_req = file("dogfood.txt", REQUIREMENTS_PATH, is_static=True)

# Templates
mit_license = file("MIT_LICENSE", TEMPLATES_PATH, "LICENSE")
setup_py = file("setup.txt", TEMPLATES_PATH, "setup.py")
rename_wheels = file("rename_testpypi_wheels.txt", TEMPLATES_PATH, "rename_testpypi_wheels.py")
dockerfile = file("Dockerfile", TEMPLATES_PATH)
makefile = file("Makefile", TEMPLATES_PATH)
makefile_docker = file("makefile.docker", TEMPLATES_PATH)
mlproject = file("MLproject", TEMPLATES_PATH)
readme = file("README.md", TEMPLATES_PATH)
contributing = file("CONTRIBUTING.md", TEMPLATES_PATH)
pyproject_toml = file("pyproject.toml", TEMPLATES_PATH)

# Workflows
push_python_wkf = file("push-python.yml", WORKFLOWS_PATH, "push.yml")
push_dist_wkf = file("push-dist.yml", WORKFLOWS_PATH, "push.yml")

ROOT_PATH_FILES = [
    gitignore,
    pyproject_toml,
    makefile,
    dockerfile,
    readme,
    contributing,
    code_of_conduct,
]

OPEN_SOURCE_FILES = [dco, mit_license]

WORKFLOW_FILES = [push_dist_wkf, push_python_wkf]

PYTHON_FILES = [init, main, version]

REQUIREMENTS_FILES = [
    lint_req,
    pytorch_req,
    tensorflow_req,
    data_viz_req,
    data_science_req,
    lint_req,
    test_req,
]

SCRIPTS = [build_manylinux_sh, dockerfile_aarch64, makefile_docker, rename_wheels]

ALL_FILES = ROOT_PATH_FILES + WORKFLOW_FILES + PYTHON_FILES + [mloq_yml]

ALL_FILE_PATHS = [str(f.src) for f in ALL_FILES]



from importlib.machinery import SourceFileLoader
from pathlib import Path

from setuptools import find_packages, setup


version = SourceFileLoader(
    "mloq.version",
    str(Path(__file__).parent / "mloq" / "version.py"),
).load_module()

with open(Path(__file__).with_name("README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="mloq",
    description="Package for initializing ML projects following ML Ops best practices.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    version=version.__version__,
    license="MIT",
    author="FragileTech",
    author_email="guillem@fragile.tech",
    url="https://github.com/FragileTech/ml-ops-quickstart",
    download_url="https://github.com/FragileTech/ml-ops-quickstart/releases",
    keywords=["Machine learning", "artificial intelligence"],
    tests_require=["pytest>=5.3.5", "hypothesis>=5.6.0"],
    install_requires=[
        "flogging>=0.0.8",
        "jinja2>=2.0.0",
        "click>=7.1.2,<8.0.0",
        "invoke>=1.4.1",
        "hydra-core>=1.0,<1.1",
    ],
    package_data={
        "": ["README.md"],
        "mloq": [
            "assets/**/*",
            "assets/**/.*",
        ],
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points="""
            [console_scripts]
            mloq=mloq.cli.main:cli
        """,
)
